from django.contrib.auth.models import Group
from django.core.exceptions import PermissionDenied

from rest_framework import viewsets

from ..models import GroupAPIKey, TrainingSet
from ..serializers import GroupSerializer
from ..permissions import VerifyGroupKey
from .utils import get_key


class GroupViewSet(viewsets.ModelViewSet):
    """
    Defines a view set for the Group module.
    Inherits from `viewsets.ModelViewSet` and defines queryset
    and serializers.
    """

    permission_classes = [VerifyGroupKey]
    serializer_class = GroupSerializer
    http_method_names = ["get"]

    def get_queryset(self):
        """
        Defining custom queryset

        :param self: A handle to the :class:`GroupViewSet`
        :type self: class

        :return: (filtered) queryset
        :rtype: QuerySet
        """
        if getattr(self, "swagger_fake_view", False):
            return TrainingSet.objects.none()
        key = get_key(self.request)
        if not key:
            raise PermissionDenied()
        try:
            api_key_object = GroupAPIKey.objects.get_from_key(key)
        except GroupAPIKey.DoesNotExist:
            raise PermissionDenied()
        if not api_key_object:
            raise PermissionDenied()
        queryset = Group.objects.filter(id=api_key_object.organization_id)
        return queryset



from rest_framework import viewsets
from rest_framework.authentication import SessionAuthentication, BasicAuthentication

from ..models import Document
from ..serializers import DocumentSerializer


class WordViewSet(viewsets.ModelViewSet):
    """
    Retrieve the list of all default words/documents or get a single record by appending the id of the requested word.
    """

    serializer_class = DocumentSerializer
    http_method_names = ["get"]
    authentication_classes = [SessionAuthentication, BasicAuthentication]

    def get_queryset(self):
        """
        Get the queryset of words/documents

        :return: The queryset of words
        :rtype: ~django.db.models.query.QuerySet
        """
        if getattr(self, "swagger_fake_view", False):
            return Document.objects.none()
        return (
            Document.objects.filter(
                creator_is_admin=True,
                document_image__confirmed=True,
                training_sets__released=True,
                training_sets__discipline__released=True,
            )
            .select_related()
            .distinct()
        )



"""
Lunes CMS URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
https://docs.djangoproject.com/en/2.2/topics/http/urls/

Examples:

Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')

Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')

Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))

"""
from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.urls import path, include
from django.conf.urls.static import static
from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.conf.urls import url
from django.templatetags.static import static as get_static_url
from django.urls import path, reverse_lazy
from django.utils.translation import ugettext_lazy as _
from django.views.generic.base import RedirectView
from django.views.i18n import JavaScriptCatalog


#: The url patterns of this module (see :doc:`topics/http/urls`)
urlpatterns = [
    path("", RedirectView.as_view(url=reverse_lazy("admin:login"))),
    path(
        "favicon.ico",
        RedirectView.as_view(url=get_static_url("images/logo.svg")),
    ),
    path("", include("lunes_cms.cms.urls")),
    url(r"^i18n/", include("django.conf.urls.i18n")),
]

urlpatterns += i18n_patterns(
    path(
        "admin/password_reset/",
        auth_views.PasswordResetView.as_view(),
        name="admin_password_reset",
    ),
    path(
        "admin/password_reset/done/",
        auth_views.PasswordResetDoneView.as_view(),
        name="password_reset_done",
    ),
    path(
        "reset/<uidb64>/<token>/",
        auth_views.PasswordResetConfirmView.as_view(),
        name="password_reset_confirm",
    ),
    path(
        "reset/done/",
        auth_views.PasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
    url(r"^admin/", admin.site.urls),
    path("i18n.js", JavaScriptCatalog.as_view(), name="javascript-translations"),
)

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

# Set dashboard title
admin.site.index_title = _("Dashboard")



"""
Map paths to view functions.
Defines custom schema views and a router that
handles the url patterns descriped in the `README.md` file
"""
from django.urls import include, path
from rest_framework import routers
from django.conf.urls import url
from rest_framework import permissions
from drf_yasg.views import get_schema_view
from drf_yasg import openapi

from . import views


class OptionalSlashRouter(routers.DefaultRouter):
    """
    Custom router to allow routes with and without trailing slash to work without redirects
    """

    def __init__(self):
        super().__init__()
        self.trailing_slash = "/?"


# Schema view for swagger documentation
schema_view = get_schema_view(
    openapi.Info(
        title="Snippets API",
        default_version="v1",
        description="Test description",
        terms_of_service="https://www.google.com/policies/terms/",
        contact=openapi.Contact(email="contact@snippets.local"),
        license=openapi.License(name="BSD License"),
    ),
    public=True,
    permission_classes=(permissions.AllowAny,),
)
# Router for dynamic url patterns
router = OptionalSlashRouter()
router.register(r"disciplines", views.DisciplineViewSet, "disciplines")
router.register(
    r"disciplines_by_level(?:/(?P<discipline_id>[0-9]+))?",
    views.DisciplineFilteredViewSet,
    "disciplines_by_level",
)
router.register(
    r"disciplines_by_group/(?P<group_id>[0-9]+)",
    views.DisciplineFilteredViewSet,
    "discipline_by_group",
)
router.register(
    r"training_set/(?P<discipline_id>[0-9]+)", views.TrainingSetViewSet, "training_set"
)
router.register(
    r"training_sets",
    views.TrainingSetByIdViewSet,
    "training_sets",
)
router.register(
    r"documents/(?P<training_set_id>[0-9]+)", views.DocumentViewSet, "documents"
)

router.register(
    r"document_by_id/(?P<document_id>[0-9]+)", views.DocumentByIdViewSet, "documents"
)
router.register(r"words", views.WordViewSet, "words")
router.register(r"group_info", views.GroupViewSet, "group_by_id")

urlpatterns = [
    path("", views.redirect_view, name="redirect"),
    path("public_upload", views.public_upload, name="public_upload"),
    path("api/", include(router.urls)),
    url(
        r"^swagger(?P<format>\.json|\.yaml)$",
        schema_view.without_ui(cache_timeout=0),
        name="schema-json",
    ),
    url(
        r"^swagger/$",
        schema_view.with_ui("swagger", cache_timeout=0),
        name="schema-swagger-ui",
    ),
    url(
        r"^redoc/$", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"
    ),
]



from django.contrib.auth.models import Group
from django.templatetags.static import static
from rest_framework import serializers

from .models import Discipline, TrainingSet, Document, AlternativeWord, DocumentImage
from .utils import get_child_count


class FallbackIconSerializer(serializers.ModelSerializer):
    """
    Serializer for models with an icon field which should provide a fallback value
    """

    icon = serializers.SerializerMethodField()

    def get_icon(self, obj):
        """
        Get the the icon if it exists and a fallback image otherwise

        :param obj: The model instance
        :type obj: ~django.db.models.Model

        :return: The url to the icon
        :rtype: str
        """
        if obj.icon:
            url = obj.icon.url
        else:
            url = static("images/fallback-icon.svg")
        request = self.context.get("request")
        if request:
            return request.build_absolute_uri(url)
        return url


class DisciplineSerializer(FallbackIconSerializer):
    """
    Serializer for the Discipline module. Inherits from
    `FallbackIconSerializer`.
    """

    total_training_sets = serializers.IntegerField()
    total_discipline_children = serializers.SerializerMethodField(
        "get_total_discipline_children"
    )
    nested_training_sets = serializers.ListSerializer(
        child=serializers.IntegerField(), source="get_nested_training_sets"
    )

    class Meta:
        """
        Define model and the corresponding fields
        """

        model = Discipline
        fields = (
            "id",
            "title",
            "description",
            "icon",
            "created_by",
            "total_training_sets",
            "total_discipline_children",
            "nested_training_sets",
        )

    def get_total_discipline_children(self, obj):
        """Returns the total child count by calling
        utils.get_child_count(obj).


        :param disc: Discipline instance
        :type disc: models.Discipline
        :return: sum of children
        :rtype: int
        """
        return get_child_count(obj)


class TrainingSetSerializer(FallbackIconSerializer):
    """
    Serializer for the TrainingSet module. Inherits from
    `FallbackIconSerializer`.
    """

    total_documents = serializers.IntegerField()

    class Meta:
        """
        Define model and the corresponding fields
        """

        model = TrainingSet
        fields = ("id", "title", "description", "icon", "total_documents")


class AlternativeWordSerializer(serializers.ModelSerializer):
    """
    Serializer for the AlternativeWord module. Inherits from
    `serializers.ModelSerializer`.
    """

    class Meta:
        """
        Define model and the corresponding fields
        """

        model = AlternativeWord
        fields = ("alt_word", "article")


class DocumentImageListSerializer(serializers.ListSerializer):
    """
    List Serializer for the DocumentImage module. Inherits from
    `serializers.ListSerializer`.
    """

    def to_representation(self, data):
        """
        Overwrite django built-in function to only deliver
        approved images.

        :param data: model instance
        :type data: models.Model
        :return: serialized model data
        :rtype: dict
        """
        data = data.filter(confirmed=True)
        return super(DocumentImageListSerializer, self).to_representation(data)


class DocumentImageSerializer(serializers.ModelSerializer):
    """
    Serializer for the DocumentImage module. Inherits from
    `serializers.ModelSerializer`.
    """

    class Meta:
        """
        Define model and the corresponding fields
        """

        list_serializer_class = DocumentImageListSerializer
        model = DocumentImage
        fields = ("id", "image")


class DocumentSerializer(serializers.ModelSerializer):
    """
    Serializer for the Document module. Inherits from
    `serializers.ModelSerializer`.
    """

    alternatives = AlternativeWordSerializer(many=True, read_only=True)
    document_image = DocumentImageSerializer(many=True, read_only=True)

    class Meta:
        """
        Define model and the corresponding fields
        """

        model = Document
        fields = (
            "id",
            "word",
            "article",
            "audio",
            "word_type",
            "alternatives",
            "document_image",
        )


class GroupSerializer(serializers.ModelSerializer):
    """
    Serializer for the Group module. Inherits from
    `serializers.ModelSerializer`.
    """

    total_discipline_children = serializers.SerializerMethodField(
        "get_total_discipline_children"
    )

    class Meta:
        """
        Define model and the corresponding fields
        """

        model = Group
        fields = (
            "id",
            "name",
            "icon",
            "total_discipline_children",
        )

    def get_total_discipline_children(self, obj):
        """Returns the total child count of a group.
        A child itself or one of its sub-children needs to
        contain at least one training set.

        :param disc: Discipline instance
        :type disc: models.Discipline
        :return: sum of children
        :rtype: int
        """
        queryset = Discipline.objects.filter(released=True, created_by=obj.id)
        queryset = [obj for obj in queryset if obj.is_root_node() and obj.is_valid()]
        return len(queryset)



from .discipline_filtered_viewset import DisciplineFilteredViewSet
from .discipline_viewset import DisciplineViewSet
from .document_by_id_viewset import DocumentByIdViewSet
from .document_viewset import DocumentViewSet
from .group_viewset import GroupViewSet
from .training_set_viewset import TrainingSetViewSet
from .training_set_by_id_viewset import TrainingSetByIdViewSet
from .other_viewset import public_upload, redirect_view
from .word_viewset import WordViewSet

__all__ = [
    "DisciplineFilteredViewSet",
    "DisciplineViewSet",
    "DocumentByIdViewSet",
    "DocumentViewSet",
    "GroupViewSet",
    "TrainingSetViewSet",
    "TrainingSetByIdViewSet",
    "public_upload",
    "redirect_view",
]



from .discipline_admin import DisciplineAdmin
from .training_set_admin import TrainingSetAdmin
from .document_admin import DocumentAdmin
from .group_api_key_admin import GroupAPIKeyAdmin

__all__ = ["DisciplineAdmin", "TrainingSetAdmin", "DocumentAdmin", "GroupAPIKeyAdmin"]



from django.contrib.auth.models import Group
from django.db import models
from django.utils.translation import ugettext_lazy as _

from rest_framework_api_key.crypto import KeyGenerator
from rest_framework_api_key.models import AbstractAPIKey, BaseAPIKeyManager

from ..utils import get_random_key


class CustomKeyGenerator(KeyGenerator):
    """
    Custom KeyGenerator that creates custom api keys, e.g. with a specific length or excluded characters.
    """

    def __init__(
        self,
        prefix_length: int,
        secret_key_length: int,
        excluded_chars: list = [],
    ):
        """Initialize KeyGenerator

        :param prefix_length: length of prefix, defaults to 3
        :type prefix_length: int, optional
        :param secret_key_length: length of secret, defaults to 7
        :type secret_key_length: int, optional
        :param excluded_chars: list of characters that should be excluded (mixed dtypes are possible), defaults to []
        :type excluded_chars: list, optional
        """
        super().__init__(
            prefix_length=prefix_length, secret_key_length=secret_key_length
        )
        self.excluded_chars = excluded_chars

    def get_prefix(self) -> str:
        """Generate prefix

        :return: Key prefix
        :rtype: str
        """
        return get_random_key(self.prefix_length, self.excluded_chars)

    def get_secret_key(self) -> str:
        """Generate secret

        :return: Key secret
        :rtype: str
        """
        return get_random_key(self.secret_key_length, self.excluded_chars)


class GroupAPIKeyManager(BaseAPIKeyManager):
    """
    Custom APIKeyManager that creates custom api keys, e.g. with a specific length or excluded characters.
    """

    key_generator = CustomKeyGenerator(
        prefix_length=3, secret_key_length=7, excluded_chars=[1, 0, "I", "l", "o", "O"]
    )


class GroupAPIKey(AbstractAPIKey):
    """
    Model that handles api keys associated with a user group.
    """

    objects = GroupAPIKeyManager()
    organization = models.ForeignKey(
        Group,
        on_delete=models.CASCADE,
        related_name="api_keys",
    )

    def __str__(self):
        return self.name + ": Token für die Organisation " + str(self.organization)

    class Meta:
        """
        Define user readable name of Document
        """

        verbose_name = _("API Key")
        verbose_name_plural = _("API Keys")



from rest_framework_api_key.admin import APIKeyModelAdmin


class GroupAPIKeyAdmin(APIKeyModelAdmin):
    exclude = ("objects",)
    list_display = [*APIKeyModelAdmin.list_display, "organization"]
    search_fields = [*APIKeyModelAdmin.search_fields, "organization"]
    list_per_page = 25
